import json
import pytz

from anytree import Node, RenderTree
from anytree.search import find

from datetime import datetime

class TaxonomyParser:
    """
    This class is a wrapper on a Tree class from the anytree library to hold
    different data hierarchies read from a JSON representation

    Source: https://towardsdatascience.com/represent-hierarchical-data-in-python-cd36ada5c71a
    """

    def __init__(self, level_prefix = "L"):
        self.prefix = level_prefix
        self.nodes = {}
        self.root_key = None

    def find_by_name(self, name) -> Node:
        """
        Retrieve a node by its unique identifier name
        """
        root = self.nodes[self.root_key]
        node = find(root, lambda node: node.name == name)
        return node
   
    def read_from_json(self, fname):
        """
        Read the taxonomy from a JSON file given as input
        """
        
        self.nodes = {}
        try:
            with open(fname, "r") as f:
                data = json.load(f)
                n_levels = len(list(data.keys()))

                # read the root node
                root = data[f"{self.prefix}0"][0]
                name = root["name"]
                _ = root.pop("name")
                
                self.nodes[name] = Node(name, **root)
                self.root_key = name

                # populate the tree
                for k in range(1, n_levels):
                    
                    key = f"{self.prefix}{k}"
                    nodes = data[key]

                    for n in nodes:
                        try:
                            assert "name" in n
                            name = n["name"]
                            _ = n.pop("name")
                            parent = n["parent"]
                            _ = n.pop("parent")
                            
                            self.nodes[name] = Node(
                                name,
                                parent=self.nodes[parent],
                                **n
                            )
                        except AssertionError:
                            print(f"Malformed node representation: {n}")
                        except KeyError:
                            print(f"Detected a dangling node: {n['name']}")

        except (FileNotFoundError, KeyError):
            raise Exception("Not existent or malformed input JSON file")

    def read_from_string(self, data_str):
        """
        Read the taxonomy from a JSON file given as input
        """
        
        self.nodes = {}
        try:
            data = json.loads(data_str)
            n_levels = len(list(data.keys()))

            # read the root node
            root = data[f"{self.prefix}0"][0]
            name = root["name"]
            _ = root.pop("name")
            
            self.nodes[name] = Node(name, **root)
            self.root_key = name

            # populate the tree
            for k in range(1, n_levels):
                
                key = f"{self.prefix}{k}"
                nodes = data[key]

                for n in nodes:
                    try:
                        assert "name" in n
                        name = n["name"]
                        _ = n.pop("name")
                        parent = n["parent"]
                        _ = n.pop("parent")
                        
                        self.nodes[name] = Node(
                            name,
                            parent=self.nodes[parent],
                            **n
                        )
                    except AssertionError:
                        print(f"Malformed node representation: {n}")
                    except KeyError:
                        print(f"Detected a dangling node: {n['name']}")

        except (KeyError):
            raise Exception("Malformed input JSON string")


class Struct:
    # https://stackoverflow.com/questions/6866600/how-to-parse-read-a-yaml-file-into-a-python-object
    def __init__(self, **entries):
        self.__dict__.update(entries)


def make_unique_task_id(nodes, datetime_stamp):
    # Create pretty-unique hash value based on input nodes
    # https://www.geeksforgeeks.org/ways-sort-list-dictionaries-values-python-using-lambda-function/
    sorted_nodes = sorted(nodes, key = lambda i: i['id'])
    nodes_str = json.dumps(sorted_nodes, sort_keys=True, ensure_ascii=True, default=str)
    nodes_hash = hashlib.sha256(nodes_str.encode('utf-8')).hexdigest()
    print(nodes_hash)
    trunc_nodes_hash = str(nodes_hash)[:8]
    task_id = f"{datetime_stamp}-{trunc_nodes_hash}"
    return(task_id, trunc_nodes_hash)

def publish_to_pubsub_topic(publisher, project_id,  topic, message):
    """Convert dictionary to JSON and publish to Pub/Sub topic.

    Args:
        publisher (pubsub.PublisherClient): Pub/Sub client
        project_id (str): Google Cloud Project ID
        topic (str): Pub/Sub topic name
        message (dict): Dictionary with header and body fields.

    Returns:
        result (???)
    """

    topic_path = publisher.topic_path(project_id, topic)
    # https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable/36142844#36142844
    json_message = json.dumps(message, indent=4, sort_keys=True, default=str).encode('utf-8')
    result = publisher.publish(topic_path, data=json_message).result()
    return result

def publish_str_to_topic(publisher, project_id, topic, str_data):
    topic_path = publisher.topic_path(project_id, topic)
    message = str_data.encode('utf-8')
    result = publisher.publish(topic_path, data=message).result()
    return result

def convert_timestamp_to_rfc_3339(timestamp):
    # Load time in RFC 3339 format
    # Description of RFC 3339: http://henry.precheur.org/python/rfc3339.html
    # Pub/Sub message example: https://cloud.google.com/functions/docs/writing/background#functions-writing-background-hello-pubsub-python
    try:
        rfc3339_time = datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%fZ')
    except ValueError as exception:
        try:
            rfc3339_time = datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%SZ')
        except:
            return ValueError
    except:
        return ValueError
    return rfc3339_time

def get_seconds_from_epoch(datetime_obj):
    """Get datetime as total seconds from epoch.

    Provides datetime in easily sortable format

    Args:
        datetime_obj (datetime): Datetime.
    Returns:
        (float): Seconds from epoch
    """
    from_epoch = datetime_obj - datetime(1970, 1, 1, tzinfo=pytz.UTC)
    from_epoch_seconds = from_epoch.total_seconds()
    return from_epoch_seconds


from .database_trigger import TriggerController
from .database_trigger import DatabaseTrigger

#from .database_trigger import Node
#from .database_trigger import Relationship

from .database_query import DatabaseQuery

from .messages import QueryRequestWriter
from .messages import QueryRequestReader
from .messages import QueryResponseWriter
from .messages import QueryResponseReader
#from .messages import JobLauncherResponse

from trellisdata import utils



import yaml

class DatabaseQuery(yaml.YAMLObject):
	"""A parameterized Neo4j query.

	Use query parameters whenver possible: https://medium.com/neo4j/neo4j-driver-best-practices-dfa70cf5a763

	args:
		name (str): Name used to request the query.
		query (str): Cypher-formatted Neo4j query.
		parameters (dict): Key-value parameters passed to the query.
		write_transaction (bool): Indicates whether this is a read or write transaction.
		active (bool): Use to activate/deactivate the query.
	"""

	# Create objects from YAML: https://pyyaml.org/wiki/PyYAMLDocumentation
	yaml_tag = u'!DatabaseQuery'

	def __init__(
				 self,
				 name,
				 query,
				 write_transaction,
				 publish_to,
				 returns = None,
				 parameters = {},
				 active = True):
		self.name = name
		self.query = query
		self.write_transaction = write_transaction
		self.publish_to = publish_to
		self.returns = returns
		self.parameters = parameters
		self.active = active



import json
import neo4j
import base64


class MessageWriter(object):

    def __init__(
                 self,
                 sender,
                 seed_id,
                 previous_event_id):

        self.message_kind = "unspecifiedWriter"
        self.sender = sender
        self.seed_id = seed_id
        self.previous_event_id = previous_event_id

    def format_json_header(self):

        message = {
           "header": {
                      "messageKind": self.message_kind,
                      "sender": self.sender,
                      "seedId": self.seed_id,
                      "previousEventId": self.previous_event_id,
           }
        }
        return message


class QueryRequestWriter(MessageWriter):

    def __init__(
                 self,
                 *, # Begin keyword-only arguments
                 sender,
                 seed_id,
                 previous_event_id,
                 query_name,
                 query_parameters,
                 custom=False,
                 query=None,
                 write_transaction=False,
                 publish_to=None,
                 returns={}):

        super().__init__(
                         sender,
                         seed_id,
                         previous_event_id)

        self.message_kind = "queryRequest"
        self.query_name = query_name
        self.query_parameters = query_parameters
        self.custom = custom
        self.query = query
        self.write_transaction = write_transaction
        self.publish_to = publish_to
        self.returns = returns

    def format_json_message(self):
        message = super().format_json_header()
        body = {
                "body": {
                    "queryName": self.query_name,
                    "queryParameters": self.query_parameters,
                    # Adding support for custom queries
                    "custom": self.custom
                }
        }
        message.update(body)

        if self.custom:
            custom_fields = {
                "query": self.query,
                "writeTransaction": self.write_transaction,
                "publishTo": self.publish_to,
                "returns": self.returns
            }
            message['body'].update(custom_fields)
        return message


class QueryResponseWriter(MessageWriter):

    def __init__(
                 self,
                 *,
                 sender,
                 seed_id,
                 previous_event_id,
                 query_name,
                 result_summary,
                 graph):

        super().__init__(
                         sender,
                         seed_id,
                         previous_event_id)

        self.message_kind = "queryResponse"
        self.query_name = query_name
        self.result_summary = result_summary
        self.graph = graph
        self.nodes = []
        self.relationships = []

        if self.graph:
            for node in self.graph.nodes:
                self.nodes.append(node)

            for relationship in self.graph.relationships:
                self.relationships.append(relationship)

    def format_json_message(self):
        message = super().format_json_header()

        body = {
           "body": {
                    "queryName": self.query_name,
                    "nodes": [self._get_node_dict(node) for node in self.nodes],
                    "relationships": [self._get_relationship_dict(rel) for rel in self.relationships],
                    "resultSummary": self._get_result_summary_dict(self.result_summary)
           }
        }
        message.update(body)
        return message

    def format_json_message_iter(self):
        for entity in self.nodes + self.relationships:
            yield self._format_json_message_single_result(entity)

    def _format_json_message_single_result(self, entity):
        result_nodes = None
        result_relationship = None

        if isinstance(entity, neo4j.graph.Node):
            node = [entity]
            relationship = []
        elif isinstance(entity, neo4j.graph.Relationship):
            node = []
            relationship = [entity]

        result_summary_dict = self._get_result_summary_dict(self.result_summary)

        message = {
           "header": {
                      "messageKind": self.message_kind,
                      "sender": self.sender,
                      "seedId": self.seed_id,
                      "previousEventId": self.previous_event_id,
           },
           "body": {
                    "queryName": self.query_name,
                    "nodes": node,
                    "relationships": relationship,
                    "resultSummary": self._get_result_summary_dict(self.result_summary)
           }
        }
        return message

    def _get_result_summary_dict(self, result_summary):
        # Create a copy of the dict so that metadata and server
        # elements are preserved in self.result_summary.
        # ResultSummary source: https://github.com/neo4j/neo4j-python-driver/blob/4.4/neo4j/work/summary.py
        summary_dict = dict(self.result_summary.__dict__)
        # The metadata field contains the dictionary of information 
        # that is parsed into attributes of the ResultSummary object.
        del summary_dict['metadata']
        # The server value is an instance of the Server class and
        # its attributes contain more classes which would require
        # parsing to text and I'm not that concerned with the
        # information anyway.
        del summary_dict['server']

        summary_dict['counters'] = summary_dict['counters'].__dict__

        return summary_dict

    def _get_relationship_dict(self, relationship):
        """Convert neo4j.graph.Relationship object into dictionary.

        neo4j.graph.Relationship source: https://github.com/neo4j/neo4j-python-driver/blob/4.4/neo4j/graph/__init__.py

        Args: 
            relationship (neo4j.graph.Relationship): Object with relationship metadata.

        Returns: 
            relationship_dict (dict): Dictionary of metadata stored in Relationship object.
        """

        relationship_dict = {
            "id": relationship.id,
            "start_node": self._get_node_dict(relationship.start_node),
            "end_node": self._get_node_dict(relationship.end_node),
            "type": relationship.type,
            "properties": dict(relationship.items())
        }
        return relationship_dict

    def _get_node_dict(self, node):
        """Convert neo4j.graph.Node object into a dictionary.

        neo4j.graph.Node source: https://github.com/neo4j/neo4j-python-driver/blob/4.4/neo4j/graph/__init__.py

        Args:
            node (neo4j.graph.Node): Object with node metadata.

        Returns:
            node_dict (dict): Dictionary of metadata stored in the Node object.
        """

        node_dict = {
            "id": node.id,
            "labels": list(node.labels),
            "properties": dict(node.items())

        }
        return node_dict


class MessageReader(object):

    def __init__(
                 self, 
                 context, 
                 event):

        self.context = context

        pubsub_message = base64.b64decode(event['data']).decode('utf-8')
        data = json.loads(pubsub_message)

        # Not sure what cases caused me to add this to db-query,
        # but keeping it in because I don't see any harm
        if type(data) == str:
            logging.warn("Message data not correctly loaded as JSON. " +
                         "Used eval to convert from string.")
            data = eval(data)

        seed_id = int(data['header'].get('seedId'))
        previous_event_id = int(data['header'].get('previousEventId'))
        if not previous_event_id:
            previous_event_id = seed_id

        self.context = context
        self.message_kind = data['header']['messageKind']
        self.header = data['header']
        self.body = data['body']
        self.sender = data['header']['sender']
        self.event_id = context.event_id
        self.seed_id = seed_id
        self.previous_event_id = previous_event_id


class QueryRequestReader(MessageReader):

    def __init__(
                 self,
                 context,
                 event):

        super().__init__(context, event)

        if self.message_kind != 'queryRequest':
            return ValueError

        self.query_name = self.body['queryName']
        self.query_parameters = self.body['queryParameters']
        self.custom = bool(self.body['custom'])

        # Should only be populated for custom queries
        if self.custom:
            self.query = self.body['query']
            self.write_transaction = bool(self.body['writeTransaction'])
            self.publish_to = self.body.get('publishTo')
            self.returns = self.body.get('returns')


class QueryResponseReader(MessageReader):

    def __init__(
                 self,
                 context,
                 event):

        super().__init__(context, event)
        
        if self.message_kind != 'queryResponse':
            return ValueError

        self.query_name = self.body['queryName']
        self.result_summary = self.body['resultSummary']
        self.nodes = self.body['nodes']
        self.relationships = self.body['relationships']



"""
class JobLauncherResponse(TrellisMessage):

    def __init__(self,
                 sender,
                 seed_id,
                 previous_event_id,
                 *,
                 job_properties):
        
        super().__init__(
                         message_kind = "jobLauncherResponse",
                         sender = sender,
                         seed_id = seed_id,
                         previous_event_id = previous_event_id)
        
        self.job_properties = job_properties

    def parse_pubsub_message(self, event, context):
        super().parse_pubsub_message(event, context)

        if self.header['messageKind'] != "jobLauncherResponse":
            return ValueError

        # Required fields
        self.job_properties = self.body['jobProperties']
    
    def format_json_message(self):
        message = {
           "header": {
                      "messageKind": self.message_kind,
                      "sender": self.sender,
                      "seedId": self.seed_id,
                      "previousEventId": self.previous_event_id,
           },
           "body": {
                    "jobProperties": self.job_properties
           }
        }
        return message
"""


import json
import yaml
import neo4j
import base64

class TriggerController:

    def __init__(self, trigger_document):
        """Manage database trigger activation and get query parameters.

        Inspired by docs: https://pyyaml.org/wiki/PyYAMLDocumentation

        args:
            trigger_document (str): YAML formatted str with tagged !DatabaseTrigger documents
        """
        triggers = yaml.load_all(trigger_document, Loader=yaml.FullLoader)

        self.node_triggers = {}
        self.relationship_triggers = {}

        # Separate triggers by type
        for trigger in triggers:
            if trigger.type == 'node':
                if trigger.start['label'] in self.node_triggers.keys():
                    self.node_triggers[trigger.start['label']].append(trigger)
                else:
                    self.node_triggers[trigger.start['label']] = [trigger]
            elif trigger.type == 'relationship':
                if trigger.start['label'] in self.relationship_triggers.keys():
                    self.relationship_triggers[trigger.start['label']].append(trigger)
                else:
                    self.relationship_triggers[trigger.start['label']] = [trigger]
            else:
                return ValueError(f"{trigger.type} is not a supported trigger type.")

    def evaluate_trigger_conditions(self, query_response):
        """

        args:
            query_response (trellisdata.QueryResponseReader)
        """

        trigger_type = self._determine_trigger_type(query_response)

        if trigger_type == 'node':
            activated_triggers = self._evaluate_node_triggers(query_response)
        elif trigger_type == 'relationship':
            activated_triggers = self._evaluate_relationship_triggers(query_response)
        else:
            return ValueError()
        return activated_triggers

    def _determine_trigger_type(self, query_response):
        len_nodes = len(query_response.nodes)
        len_relationships = len(query_response.relationships)

        if len_nodes == 1 and len_relationships == 0:
            return "node"
        elif len_nodes == 2 and len_relationships == 1:
            return "relationship"
        else:
            return ValueError()

    def _evaluate_node_triggers(self, query_response):
        node = query_response.nodes[0]
        node_label = node['labels']

        relevant_triggers = self.node_triggers[node_label]

        activated_triggers = []
        for trigger in relevant_triggers:
            node_properties = node['properties']

            query_parameters = _get_node_trigger_parameters(
                                                            trigger,
                                                            node_properties)
            trigger_tuple = (trigger, query_parameters)
            activated_triggers = append(trigger_tuple)
        return activated_triggers

    def _evaluate_relationship_triggers(self, query_response):
        relationship = query_response.relationships[0]

        start_labels = relationship['start_node']['labels']
        if len(start_labels) > 1:
            # TODO: raise warning
            pass
        start_label = start_labels[0]

        # Get relevant triggers
        relevant_triggers = self.relationship_triggers[start_label]

        end_labels = relationship['end_node']['labels']
        if len(end_labels) > 1:
            # TODO: raise warning
            pass
        end_label = end_labels[0]

        relationship_type = relationship['type']

        activated_triggers = []
        for trigger in relevant_triggers:
            if (start_label == trigger.start['label'] 
                    and end_label == trigger.end['label'] 
                    and relationship_type == trigger.relationship['type']):

                start_properties = relationship['start_node']['properties']
                end_properties = relationship['end_node']['properties']
                rel_properties = relationship['properties']

                # TODO: Get query parameters for relationship triggers
                parameters = self._get_relationship_trigger_parameters(
                    trigger,
                    start_properties,
                    end_properties,
                    rel_properties)
                trigger_tuple = (trigger, parameters)
                activated_triggers.append(trigger_tuple)
        return activated_triggers

    def _get_relationship_trigger_parameters(
                                             self,
                                             trigger,
                                             start_properties,
                                             end_properties,
                                             rel_properties):

        start_parameters = trigger.start.get('properties')
        end_parameters = trigger.end.get('properties')
        rel_parameters = trigger.relationship.get('properties')

        parameters = {}
        if start_parameters:
            parameters = self._get_parameter_values(start_parameters, start_properties, parameters)
        if end_parameters:
            parameters = self._get_parameter_values(end_parameters, end_properties, parameters)
        if rel_parameters:
            parameters = self._get_parameter_values(rel_parameters, rel_properites, parameters)
        return parameters

    def _get_node_trigger_parameters(
                                     self,
                                     trigger,
                                     node_properties):
        node_parameters = trigger.start.get('properties')

        parameters = {}
        if start_parameters:
            parameters = self._get_parameter_values(start_parameters, start_properties, parameters)
        return parameters

    def _get_parameter_values(self, parameter_mapping, properties, parameters):
        for property_name, parameter_name in parameter_mapping.items():
            parameters[parameter_name] = properties[property_name]
        return parameters


class DatabaseTrigger(yaml.YAMLObject):
    """
    Inspired by docs: https://pyyaml.org/wiki/PyYAMLDocumentation
    """
    yaml_tag = u'!DatabaseTrigger'

    def __init__(
                 self, 
                 trigger_type, 
                 start, 
                 query, 
                 end=None, 
                 relationship=None):
        self.trigger_type = trigger_type
        self.start = start
        self.query = query

        self.end = end
        self.relationship = relationship